#include <linux/ioport.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/delay.h> 
#include <asm/delay.h>
#include <string.h>
#include "version.h"
#include "hardware.h"
#include "c_table.h"

/*
 * TODO: /proc entry showing the used port addresses
 */

static int lcd_usage=0;
static char dline=0,dcol=0;


/*
 * LCD specific commands
 */

/*
 * Set to instruction mode and write a command byte
 */
void write_command(char command)
{
  outb(3,LCD_ADDRESS+2);
  outb(command,LCD_ADDRESS);
  outb(2,LCD_ADDRESS+2);
  outb(3,LCD_ADDRESS+2);
  outb(1,LCD_ADDRESS+2);
  udelay(140);
}


/*
 * Set data mode and write a data byte
 */
void write_data(char data)
{
  outb(7,LCD_ADDRESS+2);
  outb(c_table[(unsigned char)data],LCD_ADDRESS);
  outb(6,LCD_ADDRESS+2);
  outb(7,LCD_ADDRESS+2);
  outb(5,LCD_ADDRESS+2);
  udelay(140);  
}


/*
 * Resets LCD panel and sets initial mode, i.e.
 * underline cursor, not blinking
 * one line mode
 * small font
 */
void init_lcd(void)
{
  udelay(15000);
  write_command(0x38);   /* Set 8-bit mode */
  udelay(4100);
  write_command(0x38);
  udelay(100);
  write_command(0x38);
  udelay(4100);
  write_command(0x06);  
  udelay(40);
  write_command(0x0c);
  udelay(40);
  write_command(1);    /* Set entry mode, cursor moves right, shift display */
  udelay(1640);
  write_command(2);
  udelay(1640);
}


/*
 * Following functions are for tty like handling of the panel
 * i.e. set cursor position, clear display, home cursor, etc.
 */

/*
 * Sets cursor to certain position
 * col  = column [0-19]
 * line = line [0-3]
 */
void pos_cursor(unsigned char col, unsigned char line)
{
  switch(line) {
    case 1: write_command(0x80+col); break;
    case 2: write_command(0xc0+col); break;
    case 3: write_command(0x94+col); break;
    case 4: write_command(0xd4+col); break;
    default: break;
    }
}


/*
 * All functions after here are for device driver specific purposes
 */

/*
 * Called when the device is open()ed
 */
static int lcd_open(struct inode *minode, struct file *file)
{
  if (lcd_usage!=0)
    return -EBUSY;

  MOD_INC_USE_COUNT;
  lcd_usage=1;
#ifdef DEBUG
  printk("LCD opened\n");
#endif

  return(0);
}


/*
 * Called when the device is close()ed
 */
static void lcd_release(struct inode *minode, struct file *mfile)
{
  MOD_DEC_USE_COUNT;
  lcd_usage=0;
#ifdef DEBUG
  printk("LCD module closed\n");
#endif
}


/*
 * Called if data is to be written out to the device
 */
int lcd_write_byte(struct inode *minode, struct file *mfile, const char *gdata, int length)
{
int i;
const char *data;
char c;

  data=gdata;

  for (i=0; i<length; i++) {
    c=get_user((char *)(data+i));
/*    printk("LCD: starting %d\n",i);
    printk("LCD: which is '%c'\n",c);*/
    if (c==10) { /* line feed */
      dline++;
      dcol=0;
      pos_cursor(dcol,dline);
    } else
    if (c==27) {
      if ((i+1)<length) {
        i+=1;
        c=get_user((char *)(data+i));
        if (c=='c') { 
          write_command(1);
          dcol=0;
          dline=0;
        } else
        if (c=='h') { 
          write_command(128);
          dcol=0;
          dline=0;
        } else
        if (c=='R') {
          init_lcd();
          dcol=0;
          dline=0;
        }
      }
    } else
    if (c==13) { /* carriage return */
      dcol=0;
      pos_cursor(dcol,dline);
    } else
    {
      write_data(c);
      dcol++;
      if (dcol>(LCD_COLS-1)) {
        dcol=0;
        dline++;
        pos_cursor(dcol,dline);
      }
    }
  }
  return(length);
}


/*
 * Called if data is to be read
 */
int lcd_read_byte(struct inode *inode, struct file *mfile, char *data, int length)
{
#ifdef DEBUG
  printk("LCD: Trying to read readonly device\n");
#endif
  return(EACCES);
/*
  sprintf(data,"%d %d\x004",dcol,dline);
  return(strlen(data));
*/
}


/*
 * Called when module is unloaded
 */
void cleanup_module(void)
{
  release_region(LCD_ADDRESS,3);
  printk("LCD: I/O region freed\n");
  if (unregister_chrdev(LCD_MAJOR,"LCD panel"))
    printk("LCD: module cleanup failed\n");
  else
    printk("LCD: module cleanup succeeded\n");
}


/*
 * First function called when module is inserted
 */
int init_module(void)
{
  static struct file_operations lcd_fops={NULL,lcd_read_byte,lcd_write_byte,
            NULL,NULL,NULL,NULL,lcd_open,lcd_release,
            NULL,NULL,NULL,NULL};
  if (register_chrdev(LCD_MAJOR,"LCD panel",&lcd_fops))
    printk("unable to get major 120 for LCD device\n");
  else {
    printk("LCD device %s init:\n",LCD_MODULE_VERSION);
    if (!check_region(LCD_ADDRESS,3)) {
      request_region(LCD_ADDRESS,3,"LCD panel");
      printk("  got 3 addresses from %#x\n",LCD_ADDRESS);
      lcd_usage=0;
      init_lcd();
      dline=0;
      dcol=0;
    } else
    {
      release_region(LCD_ADDRESS,3);
      unregister_chrdev(LCD_MAJOR,"LCD panel");
      printk("  could not get 3 addresses from %#x\n",LCD_ADDRESS);
    }
  }
  return 0;
}
